// $Log$
// Revision 1.13  2010/04/28 19:06:51  stuart
// Nits from 5.4 compiler
//
// Revision 1.12  2010/04/02 04:14:29  stuart
// Convert Array to vector
//
// Revision 1.11  2006/11/16 21:55:56  stuart
// Compile on CentOS-4
//
// Revision 1.10  2004/02/09 17:28:52  stuart
// Port from libg++ to STL
//
// Revision 1.9  2003/03/24 19:49:57  stuart
// Support ANSI date/time changes
//
// Revision 1.8  2001/02/20 22:05:53  stuart
// convert to bool
//
// Revision 1.7  1999/01/29  20:32:22  ed
// Support fixed point in "floating" point load routine.
//
// Revision 1.6  1997/04/18  21:37:00  stuart
// report element id when mandatory field missing.
//
// Revision 1.5  1996/04/30  19:14:46  stuart
// various bugs and typos
//
// Revision 1.4  1994/10/20  14:11:52  stuart
// numerous formatting and null element refinements
//
// Revision 1.3  1993/11/16  20:15:42  stuart
// additional validation
//
// Revision 1.2  1993/10/28  15:23:07  stuart
// supress leading blanks for N format
//
#pragma implementation
#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <Obstack.h>
#include <Btcb.h>
#include <ftype.h>
#include "edi.h"
#include "cicode.h"

#ifdef _AIX
extern "C" time_t timegm(struct tm *t) {
  const char *savtz = getenv("TZ");
  putenv("TZ=GMT");
  tzset();
  time_t ts = mktime(t);
  putenv(savtz);
  tzset();
  return ts;
}
#endif

bool CodeDict::century = false;

CodeDict::CodeDict(const char *n):
  name(n),
  rec(*new cimp_mast),
  cmf( *new btFile(Keydata(&rec,sizeof rec)))
{
  len = 512;
  items = new EDIitem *[len];
  for (int i = 0; i < len; ++i) items[i] = 0;
  mask = len - 1;
  cnt = 0;
}

CodeDict::~CodeDict() {
  delete &cmf;
  delete &rec;
  delete items;
}

void CodeDict::rehash(int newsize) {
  EDIitem** old = items;
  int oldlength = len;
  items = new EDIitem*[newsize];
  len = newsize;
  for (int i = 0; i < newsize; ++i) items[i] = 0;
  mask = newsize - 1;
  for (int i = 0; i < oldlength; ++i) {
    EDIitem *e = old[i];
    if (e == 0) continue;
    int h = e->id & mask;
    while (items[h] != 0)
      h = (h - 1) & mask;
    items[h] = e;
  }
}

EDIitem &CodeDict::operator[](int i) {
  for (int h = i & mask;;h = (h - 1) & mask) {
    EDIitem *e = items[h];
    if (e == 0) {
      // allocate new item
      if (cmf.keylen() == 0) {
	if (cmf.open(name.c_str(),BTRDONLY + NOKEY)) {
	  cerr << name << ": file not found\n";
	}
      }
      cimp_mast newrec;
      newrec.elem = i;
      if (cmf.readeq(&newrec.elem)) {
	cerr << i << ": data element not found\n";
	rec = newrec;
      }
      // patch meta data for date items when century flag is set
      if (century && rec.type == "DT") {
	rec.max = 8;
	rec.min = 8;
      }
      e = EDIitem::create(rec);

      items[h] = e;
      if (2 * ++cnt > len)
	rehash(2 * len);
      return *e;
    }
    if (e->id == i)
      return *e;
  }
}

struct tagarr {
  SegDict::Tag *p;
  char *req;
};

SegDict::SegDict(const char *filename): h(*new Obstack) {
  exp = 0;
  cache = -1;
  FILE *f = fopen(filename,"r");
  if (!f) {
    perror(filename);
    return;
  }
  int c;
  char *elemlist;
  const void *segdesc = "SEGMENT";
  for (;;) {
    switch (c = getc(f)) {
    case EOF:
      fclose(f);
      return;
    case '\t':
      name.push_back((const char *)h.finish(0));
      continue;
    case '\n':
      c = getc(f);
      if (c == '\t') {
	segdesc = h.finish(0);
	continue;
      }
      ungetc(c,f);
      elemlist = (char *)h.finish(0);
      break;
    default:
      h.grow(c);
      continue;
    }
    h.free(elemlist);	// overwrite ascii text w/ array of Tag
    Tag e;
    char req[128], *rp = req;
    for (char *elem = strtok(elemlist,","); elem; elem = strtok(0,",")) {
      e = atoTag(elem);
      h.grow(&e,sizeof e);
      char opt = 'O';
      if (*elem) {
	opt = elem[strlen(elem)-1];
	if (!isalpha(opt))
	  opt = 'O';
      }
      *rp++ = opt;
    }
    e = 0;
    h.grow(&e,sizeof e);
    tagarr t;
    t.p = (Tag *)h.finish();
    *rp = 0;
    t.req = (char *)h.copy(req);
    list.push_back((const tagarr *)h.copy(&t,sizeof t));
    segn.push_back((const char *)segdesc);
  }
}

SegDict::~SegDict() {
  delete &h;
}

const char *SegDict::operator()(const char *segname) {
  int i = find(segname);
  return (i >= 0) ? (const char *)segn[i] : "SEGMENT";
}

const char *SegDict::operator()(Tag t) {
  return (*this)(Tagtoa(t));
}

const SegDict::Tag *SegDict::elems(const char *segname) {
  int i = find(segname);
  return (i >= 0) ? ((const tagarr *)list[i])->p : 0;
}

const char *SegDict::req(const char *segname) {
  int i = find(segname);
  return (i >= 0) ? ((const tagarr *)list[i])->req : 0;
}

const SegDict::Tag *SegDict::elems(Tag t) {
  return elems(Tagtoa(t));
}

const SegDict::Tag *SegDict::operator[](const char *segname) {
  const Tag *p = elems(segname);
  if (!p) return p;
  if (exp)
    h.free(exp);
  expand(exp = (Tag *)h.alloc(size(p) * sizeof *exp),p);
  return exp;
}

int SegDict::isComposite(Tag i) {
  return i > 9999;
}

int SegDict::size(const Tag *p) {
  int n = 0;
  // compute total number of expanded elements
  if (p) for (int i = 0; p[i]; ++i) {
    if (isComposite(p[i]))
      n += size(elems(Tagtoa(p[i])));
    else
      ++n;
  }
  return n;
}

int SegDict::expand(Tag *d,const Tag *s) {
  int i = 0;
  if (s) for (; s[i]; ++i) {
    if (SegDict::isComposite(s[i]))
      i += expand(d+i,s+i);
    else
      d[i] = s[i];
  }
  d[i] = 0;
  return i;
}

int SegDict::find(const char *segname) {
  if (cache >= 0 && strcmp(segname,(const char *)name[cache]) == 0)
    return cache;
  for (size_t i = 0; i < name.size(); ++i) {
    if (strcmp(segname,(const char *)name[i]) == 0) {
      return cache = i;
    }
  }
  return cache = -1;
}

int SegDict::segcnt() const {
  return name.size();
}

SegDictLibrary SegDict::library("/bms/edi");

SegDictLibrary::SegDictLibrary(const string &d): dir(d) {
  if (dir.length() && dir != "/")
    dir += '/';
}

SegDict *SegDictLibrary::findVersion(const string &s) {
  for (size_t i = 0; i < key.size(); ++i)
    if (key[i] == s)
      return data[i];
  string name(dir + s + ".seg");
  SegDict *p = new SegDict(name.c_str());
  if (p->segcnt() == 0) {
    delete p;
    return 0;
  }
  key.push_back(s);
  data.push_back(p);
  return p;
}

SegDictLibrary::~SegDictLibrary() {
  for (size_t i = 0; i < data.size(); ++i)
    delete data[i];
}

EDIitem::EDIitem(const cimp_mast &m):
  desc(m.desc), type(m.type), min(m.min), max(m.max), id(m.elem)
{
  size = max;
  mask = 0;
}

Julday toJulday(const string &in) {
  string s(in);
  if (s.length() == 6) {	// YYMMDD
    string y(s.substr(0,2)), m(s.substr(2,2)), d(s.substr(4,2));
    int year = atoi(y.c_str());
    if (year < 70) year += 100;
    return Julday(
      atoi(m.c_str()),	// month
      atoi(d.c_str()),	// day
      year + 1900	// year
    );
  }
  else {			// CCYYMMDD
    string y(s.substr(0,4)), m(s.substr(4,2)), d(s.substr(6,2));
    return Julday(
      atoi(m.c_str()),	// month
      atoi(d.c_str()),	// day
      atoi(y.c_str())	// year
    );
  }
}

struct EDIdate: EDIitem {
  EDIdate(const cimp_mast &m);
  void store(const string &s,char *p);
  string load(const char *p);
};

EDIdate::EDIdate(const cimp_mast &m): EDIitem(m) {
  mask = 1;
  size = 4;
}

void EDIdate::store(const string &s,char *p) { stlong(toJulday(s),p); }

string EDIdate::load(const char *p) {
  return Julday(ldlong(p)).pic((max > 6)?"YYYYMMDD":"YYMMDD");
}

/** Normalize and validate the character representation of an
    ANSI numeric field. */
static int numlen(string &s) {
  int len = 0;
  if (s.length()) {
    if (s[0] == '.')
      s = "0" + s;
    if (s[s.length() - 1] == '.')
      s += '0';
  }
  for (size_t i = 0; i < s.length(); ++i) {
    int c = s[i];
    if (isdigit(c))
      ++len;
    else if (c != '-' && c != '.')
      return -1;	// invalid char
  }
  return len;
}

static const char nulval[8] = { 0x80 };

class EDInum: public EDIitem {
  short fix;
public:
  EDInum(const cimp_mast &m);
  void store(const string &s,char *p) { Money(s.c_str(),"-").store(p,size); }
  bool isnull(const char *p) { return memcmp(p,nulval,size) == 0; }
  const char *validate(string &s);
  string load(const char *p);
};

EDInum::EDInum(const cimp_mast &m): EDIitem(m) {
  if (type.length() > 1 && '0' <= type[1] && type[1] <= '9')
    fix = type[1] - '0';
  else
    fix = 0;
  mask = 1;
  if (size < 5)
    size = 2;
  else if (size < 10)
    size = 4;
  else
    size = 6;
}

string EDInum::load(const char *p) {
  return dec(Money(p,size),min);
}

static const char *numval(string &s,int min) {
  int len = numlen(s);
  if (len < 0) return "37";	// invalid type of character
  while (len < min) {
    s = "0" + s;
    len = numlen(s);
  }
  return 0;
}

const char *EDInum::validate(string &s) { return numval(s,min); }

struct EDItime: EDIitem {
  EDItime(const cimp_mast &m): EDIitem(m) {
    mask = 1;
    if (size < 5)
      size = 2;
    else
      size = 4;
  }
  void store(const string &s,char *p) {
    long t = atol(s.c_str());
    if (size == 2)
      stshort(t,p);
    else
      stlong(t,p);
  }
  bool isnull(const char *p) { return memcmp(p,nulval,size) == 0; }
  const char *validate(string &s) { return numval(s,min); }
  string load(const char *p) {
    if (size == 2)
      return dec((long)ldshort(p),4);
    return dec(ldlong(p),6);
  }
};

struct EDIflt: EDIitem {
  EDIflt(const cimp_mast &m): EDIitem(m) {
    mask = 1;
    if (size < 7)
      size = 4;
    else
      size = 8;
  }
  void store(const string &s,char *p) {
    if (size == 4)
      sts1float(atof(s.c_str()),p);
    else
      sts1extfloat(atof(s.c_str()),p);
  }
  bool isnull(const char *p) { return memcmp(p,nulval,size) == 0; }
  const char *validate(string &s);
  string load(const char *p);
};

const char *EDIflt::validate(string &s) {
  int len = numlen(s);
  if (len < 0) return "37";	// invalid type of character
  while (len < min) {
    s = "0" + s;
    len = numlen(s);
  }
  return 0;
}

string EDIflt::load(const char *p) {
  double r;
  if (p[0] >= 0 && p[0] < 6) // fixed point format
    return (size == 4)
      ? Money(p+1,3).fmt(p[0])
      : Money(p+2,6).fmt(ldshort(p));
  if (size == 4)
    r = lds1float(p);
  else
    r = lds1extfloat(p);
  char buf[40];
  return string(gcvt(r,max,buf));
}

void EDIitem::store(const string &s,char *p) {
  stebstr(s.c_str(),(unsigned char *)p,max);
}

string EDIitem::load(const char *p,bool mand,const char *elemid) {
  if (isnull(p)) {
    if (mand) {
      cerr << "Mandatory elem `";
      if (elemid)
	cerr << elemid << ": ";
      cerr << '(' << id << ") " << desc << "' missing!\n";
    }
    return "";
  }
  string s(load(p));
  return (mand || s.length()) ? trim(s,min,max) : s;
}

string EDIitem::load(const char *p) {
  char *s = (char *)alloca(max + 1);
  ldebstr(s,(unsigned char *)p,max);
  return trim(s,0,max);
}

bool EDIitem::isnull(const char *p) {
  for (int i = 0; i < size; ++i)
    if (p[i]) return 0;
  return 1;
}

const char *EDIitem::validate(const char *) {
  return 0;
}

const char *EDIitem::validate(string &s) {
  const char *err = 0;
  while ((int)s.length() < min)
    s += ' ';
  if ((int)s.length() > max)
    s.substr(max) = "";
  return err;	// default to no error
}

struct EDIid: EDIitem {
  EDIid(const cimp_mast &m): EDIitem(m) { }
  bool isnull(const char *p) { return *p == 0x40; }
  const char *validate(const char *p) {
    if (p[min-1] != 0x40) return 0;
    return "4";
  }
};

EDIitem *EDIitem::create(cimp_mast &m) {
  if (m.type[0] == 'N') {
    return new EDInum(m);
  }
  if (m.type == "R") {
    return new EDIflt(m);
  }
  if (m.type == "AN") {
    m.type = "C";
    return new EDIitem(m);
  }
  if (m.type == "DT") {
    return new EDIdate(m);
  }
  if (m.type == "TM") {
    return new EDItime(m);
  }
  if (m.type == "ID") {
    return new EDIid(m);
  }
  return new EDIitem(m);
}

time_t toDateTime(string &dt, string &tim) {
  tm t;
  toJulday(dt).mdy(t);
  long time = atol(tim.c_str());
  if (time > 9999) {
    t.tm_sec = time % 100;
    time /= 100;
  }
  t.tm_min = time % 100;
  t.tm_hour = time / 100;
  return timegm(&t);
}

string trim(string s,int min,int max) {
  int len = s.length();
  if (len < min)
    return s + string(min - len,' ');
  while (len > min && s[len - 1] == ' ')
    s.erase(--len,1);
  if (len > max)
    return s.substr(0,max);
  return s;
}
