#pragma interface
#include <string.h>
#include <Ftype.h>
#include <vector>

struct SegDictLibrary;
struct tagarr;

struct SegDict {
  SegDict(const char *);
  ~SegDict();
  typedef short Tag;			// a data element tag
  int size(const Tag *);		// number of elements when expanded
  int segcnt() const;			// number of segments
  static int isComposite(Tag);
  static SegDictLibrary library;
  const char *operator()(const char *);	// segment description
  const char *operator()(Tag t);
  const Tag *operator[](const char *);	// expanded data elements
  const Tag *elems(const char *);	// composite + simple elements
  const char *req(const char *);	// required flags
  const Tag *elems(Tag t);
  int expand(Tag *s,const Tag *);	// expand into s, return size
private:
  class Obstack &h;
  vector<const char *> name, segn;
  vector<const tagarr *>list;
  SegDict(const SegDict &);
  SegDict&operator=(const SegDict &);
  int cache;
  int find(const char *);
  Tag *exp;	// last expanded elements
};

SegDict::Tag atoTag(const char *s);
const char *Tagtoa(SegDict::Tag i);

struct SegDictLibrary {
  SegDictLibrary(const string &dir);
  SegDict *findVersion(const string &);
  ~SegDictLibrary();
private:
  SegDictLibrary(const SegDictLibrary &);
  SegDictLibrary&operator=(const SegDictLibrary &);
  string dir;
  vector<string> key;
  vector<SegDict *> data;
};

class EDIitem {
public:
  const short id;
protected:
  short mask;
  short size;
public:
  const short min,max;
  static EDIitem *create(class cimp_mast &);
  string type;
  string desc;
  EDIitem(const class cimp_mast &m);
  int align(int pos) const { return (pos + mask) & ~mask; }	// EDX align
  int len() const { return size; }		// EDX binary size
  virtual void store(const string &,char *);
  string load(const char *,bool mand,const char *elemid = 0);
  virtual string load(const char *);
  virtual bool isnull(const char *);
  virtual const char *validate(const char *);	// validate EDX data
  virtual const char *validate(string &);	// validate EDI data
  virtual ~EDIitem() {}
};

class CodeDict {
  class cimp_mast &rec;
  class btFile &cmf;
  EDIitem **items;
  int len;
  int mask;
  int cnt;
  string name;
  void rehash(int newsize);
public:
  /** Set to true to enable 4 digit years for date fields. */
  static bool century;
  CodeDict(const char *n);
  int getSize() { return cnt; }
  EDIitem &operator[](int);
  ~CodeDict();
};

string trim(string s,int min,int max);
time_t toDateTime(string &dt, string &tim);
Julday toJulday(const string &dt);
