#pragma implementation
#include <Btcb.h>
#include <Ftype.h>
#include <string.h>
#include <edifact.h>
#include <Date.h>
#include "edienv.h"
#include "edilog.h"

static msglog mlog;

EDIstream &operator<<(EDIstream &s,tm *t) {
  char buf[16];
  tm_mask(t,"YYMMDDhhmm",buf);
  s << string(buf,6) << string(buf+6,4);
  return s;
}

static void expand(char *buf,const char *s) {
  if (s) while (*s) {
    if (s[1] == '-' && s[2]) {
      for (int i = (unsigned char)s[0]; i <= (unsigned char)s[2]; ++i)
	*buf++ = i;
      s += 3;
    }
    else 
      *buf++ = *s++;
  }
  *buf = 0;
}

// define the set of valid transmission chars
void EdifactEnv::charset(const char *s) {
  // initialize map to produce telex compatible chars only
  memset(map,':',128);
  memset(map+128,'.',MAPSIZE - 128);
  // now make all the chars in the charset stand for themselves
  char fromset[MAPSIZE];
  expand(fromset,s);
  for (int i = 0; fromset[i]; ++i)
    map[(unsigned char)fromset[i]] = fromset[i];
}

void EdifactEnv::charmap(const char *s,const char *m) {
  char fromset[MAPSIZE], toset[MAPSIZE];
  expand(fromset,s);
  expand(toset,m);
  for (int i = 0; fromset[i] && toset[i]; ++i) {
    int c = (unsigned char)fromset[i];
    int to = (unsigned char)toset[i];
    if ((unsigned char)map[c] != c)
      map[c] = (c == to) ? to : map[to];
  }
}

static const char unob[] = "a-z*!;<>%&\" .,():+'=/?A-Z0-9-";
static const char unoa[]  = "*!;<>%&\" .,():+'=/?A-Z0-9-";
static const char telex[] = " .,():+'=/?A-Z0-9-";

EdifactEnv::EdifactEnv() {
  map = new char[MAPSIZE];
  syntax("UNOB",2);
  syn[4] = 0;
  priority = 0;
  grps = 0;
}

EdifactEnv::~EdifactEnv() {
  UNZ();
  delete [] map;
}

void EdifactEnv::syntax(const char *s,short release) {
  strncpy(syn,s,4);
  rel = release;
  if (strcmp(syn,"UNOB") == 0)
    charset(unob);
  else if (strcmp(syn,"UNOA") == 0)
    charset(unoa);
  else
    charset(telex);	// basic telex character set
  charmap("a-z*!;<>&\"","A-Z .,()+'");	// make mapping more readable
}

void EdifactEnv::UNB(string f, string t, bool tst) {
  now = time(0);
  grps = 0;
  test = tst;
  split(f,from,3,string(1,sep[COMP]));
  split(t,to,3,string(1,sep[COMP]));
  while (from.size() < 3) from.push_back("");
  while (to.size() < 3) to.push_back("");

  EDIstream &self = *this;
  UNZ();
  tm *today = gmtime(&now);
  char prior[2];
  prior[0] = priority;
  prior[1] = 0;
  self << "UNB"
      << ::elem << syn << (long)rel
      << ::elem << from[0] << from[1] << from[2]
      << ::elem << to[0] << to[1] << to[2]
      << ::elem << today
      << ::elem /* null */
      << ::elem << passwd << passwdQual
      << ::elem << ref
      << ::elem << prior
      << ::elem // ack request
      << ::elem << agreement;
  if (test)
    self << ::elem(11) << '1';
  mlog.log(seg);
  seq = seg(5);
  self << eseg;
}

void EdifactEnv::UNZ() {
  if (grps) {
    (*this) << "UNZ" << ::elem << grps << ::elem << seq << eseg;
    grps = 0;
    toplevel = mlog.NONE;
  }
}

// scan for invalid chars and substitute the first telex char
int EdifactEnv::write(Segment &seg) {
  // FIXME: check for UNG/UNE
  char type = mlog.log(seg);
  if (!grps) switch (type) {
  case msglog::UNH: case msglog::UNG:
    toplevel = type;
  }
  if (type == toplevel)
    ++grps;
  // map to target character set
  int n = seg.size();
  for (int i = 0; i < n; ++i) {
    int m = seg.size(i);
    for (int j = 0; j < m; ++j) {
      string &str = seg(i,j);
      int len = str.length();
      for (int k = 0; k < len; ++k) {
	int c = (unsigned char)str[k];
	if (c != map[c])
	  str[k] = map[c];
      }
    }
  }
  return EdifactSink::write(seg);
}
