#pragma implementation
#include <string.h>
#include <iostream>
#include <edifact.h>
#include <Money.h>
#include "edi.h"

static CodeDict codedict("/edx/AIRCO/EDIFACT/CODE.MAST");
static SegDict segdict("/bms/edi/edifact.seg");

Edifact::Edifact(const char *s): EdifactSink(s) { }

Edifact::~Edifact() {
  close();
}

void EdifactSink::UNA(const char *s) {
  strncpy(sep,s,sizeof sep);
  sep[sizeof sep - 1] = 0;
  needUNA = true;
}

void Edifact::MOA(int type,Money amt,bool req) {
  EDIstream &msg = *this;
  if (amt > Money(0L))
    msg << "MOA" << ::elem << type << amt.fmt(2) << eseg;
  else if (req)
    msg << "MOA" << ::elem << type << eseg;
}

EdifactSink::EdifactSink(const char *s) {
  strncpy(sep,":+,? '",sizeof sep);
  needUNA = false;
  if (s) UNA(s);
}

EdifactSink::~EdifactSink() { }

int Edifact::write(Segment &seg) {
  int n = seg.size();
  if (seg[0] == "UNH") {
    close();
    msgref = seg[1];
  }
  const SegDict::Tag *t = segdict.elems(seg[0].c_str());
  if (t == 0) {
    cerr << seg[0] << ": invalid segment name\n";
    return -1;
  }
  for (int i = 0; i < n; ++i) {
    int m = seg.size(i);
    const SegDict::Tag *s = 0;
    if (i) {
      if (t[i - 1] == 0) {
	cerr << seg << "\n*** too many elements for segment\n";
	break;
      }
      s = segdict.elems(t[i - 1]);
    }
    if (i && !s) {
      if (m > 1) {
	cerr << seg[i] << "\n*** subelements used when not a composite\n";
	m = 1;
      }
      s = &t[i - 1];
    }
    for (int j = 0; j < m; ++j) {
      string &str = seg(i,j);
      if (i) {
	if (s[j] == 0) {
	  cerr << seg << "\n*** too many subelements for composite ("
		<< i << ',' << j << ")\n";
	  break;
	}
	EDIitem &item = codedict[s[j]];
	const char *err = item.validate(str);
	if (err)
	  cerr << str << ": " << err << endl;
	if (sep[DEC] != '.' && (item.type == "N" || item.type == "R"))
	  gsub(str,'.',sep[DEC]);
      }
    }
  }
  return EdifactSink::write(seg);
}

int EdifactSink::write(Segment &seg) {
  if (needUNA) {
    needUNA = false;
    cout << "UNA" << sep;
    if (false && sep[SEG] != '\n')
      cout << endl;
  }
  int last = 0;
  int n = seg.size();
  for (int i = 0; i < n; ++i) {
    int m = seg.size(i);
    while (m && seg(i,m-1).length() == 0) --m;
    if (!m) continue;
    while (i > last) {
      cout << sep[ELEM];
      ++last;
    }
    for (int j = 0; j < m; ++j) {
      string str(seg(i,j)), esc(1,sep[ESC]);
      gsub(str,esc,esc + sep[ESC]);
      gsub(str,string(1,sep[COMP]),esc + sep[COMP]);
      gsub(str,string(1,sep[ELEM]),esc + sep[ELEM]);
      gsub(str,string(1,sep[SEG]),esc + sep[SEG]);
      if (j) cout << sep[COMP];
      cout << str;
    }
  }
  cout << sep+SEG;
  if (false && sep[SEG] != '\n')
    cout << endl;
  ++cnt;
  seg.clear();
  return 0;
}

void Edifact::close() {
  if (cnt) {
    Segment s;
    s + "UNT" + dec(++cnt) + msgref;
    write(s);
    cnt = 0;
  }
}
