#pragma implementation
#include <iostream>
#include <btflds.h>
#include "edi.h"
#include "edifile.h"

CodeDict codedict("/edx/AIRCO/EDI/CODE.MAST");

void EDIfile::error(short pos,short elem,const string &id,const string &val) {
  Segment seg;
  EDIsink::error(seg + "AK4"+dec((long)pos)+dec((long)elem)+id+val);
}

EDIfile::EDIfile(const string &name,SegDict *vers):
  fname(name), tname("TMP/")
{
  ok = 0;
  // create in work directory
  static char tbl[] = { 0, 1,
    BT_NUM, 4,
    BT_CHAR, 3,
    BT_BIN, 255, BT_BIN, 200
  };
  btflds f;
  tname += basename(name.c_str());
  btcreate(tname.c_str(),ldflds(&f,tbl,sizeof tbl),0644);
  int rc = bt.open(tname.c_str(),BTWRONLY);
  cerr << "EDIfile: " << name << ", rc = " << rc << endl;
  ok = rc == 0;
  segdict = vers;
}

EDIfile::~EDIfile() {
  // move to spool directory if OK, else delete
  if (ok)
    btrename(tname.c_str(),fname.c_str());
  else
    btkill(tname.c_str());
}

int EDIfile::write(Segment &fld) {
  const SegDict::Tag *p;
  if (!segdict) return -1;
  if (!fld.size() || !(p = (*segdict)[fld[0].c_str()])) {
    cerr << "EDIfile: " << fld[0] << ": invalid segment name\n";
    cerr << fld << endl;
    return -1;
  }
  char *lbuf = bt.lbuf();
  stlong(++seq,lbuf);
  int pos = 4;
  int flen = 3;
  stchar(fld[0].c_str(),lbuf + pos,flen);
  pos += flen;
  for (int i = 0; p[i]; pos += flen) {
    EDIitem &c = codedict[p[i++]];
    if (i >= fld.size()) break;
    flen = c.len();
    if (flen) {
      pos = c.align(pos);
      const char *err = c.validate(fld[i]);
      if (err)
	error(i,p[-1],err,fld[i]);	// report error
      c.store(fld[i],lbuf + pos);
    }
  }
  int rc =  bt.write(pos,4);
  if (rc == 0)
    ++cnt;
  return rc;
}

class TranId: btFile {
  FNUM<4> id;
  string name;
public:
  TranId(const char *n): name(n), btFile(&id) { }
  string assign();
  string cur() const { return dec(long(id)); }
};

TranId tranId("LOG");

string TranId::assign() {
  if (id == 0)
    open(name.c_str(),BTRDWR);
  if (last())
    id = 0;
  do
    ++id;
  while (write());
  return dec(long(id));
}

STfile::STfile(const string &type,const string &filename,SegDict *vers):
  EDIfile(filename+'.'+tranId.assign(),vers), id(tranId.cur())
{
  Segment seg;
  write(seg+"ST"+type+id);
}

STfile::~STfile() {
  Segment seg;
  string segcnt = dec(cnt + 1);
  write(seg +"SE"+segcnt+id);
}
