#pragma implementation
#include <iostream>
#include <stdlib.h>
#include "edi.h"
#include "partner.h"
#include "edilog.h"

int msglog::type(Segment &seg) {
  static const char *sname[] = { "",
    "ISA", "GS" , "ST" , "SE" , "GE" , "IEA",
    "UNB", "UNG", "UNH", "UNT", "UNE", "UNZ", 0
  };
  string &id = seg(0,0);
  for (int i = 1; sname[i]; ++i)
    if (id == sname[i]) return i;
  return NONE;
}

int msglog::log(Segment &seg) {
  int t = type(seg);
  switch (t) {
  case UNB:
    logUNB(seg);
    break;
  case UNZ:
    if (seg(2) == "")
      seg(2) = unbrec.control;
    break;
  case UNH:
    logUNH(seg);
    break;
  case UNT:
    if (seg(2) == "")
      seg(2) = unhrec.control;
    break;
  }
  return t;
}

msglog::msglog(): unbrec(*new UNBlog), unhrec(*new UNHlog) {
  unbf = Keydata(&unbrec,sizeof unbrec);
  unhf = Keydata(&unhrec,sizeof unhrec);
}

msglog::~msglog() {
  delete &unbrec;
  delete &unhrec;
}

int msglog::logUNH(Segment &seg) {
  if (unhf.keylen() == 0) {
    int rc = unhf;
    if (rc == 0) {
      rc = unhf.open("UNHLOG",BTRDWR + NOKEY);
      if (rc)
	cerr << "UNHLOG: BTAS/X file not found.\n";
    }
    if (rc) return rc;
  }
  unhrec.from = unbrec.from;
  Money seq;
  if (seg(1) == "") {
    char prefix = '0';
    unhf.last(&unhrec.from + Keydata(&prefix,1));
    seq = ++Money(string(unhrec.control).c_str(),"##########");
    seg(1) = seq.pic("##########");
  }
  for (;;) {
    unhrec.control = seg(1);
    unhrec.msgtype = seg(2);
    unhrec.msgver = atol(seg(2,1).c_str());
    unhrec.msgrel = atol(seg(2,2).c_str());
    unhrec.agency = seg(2,3);
    unhrec.code = seg(2,4);
    unhrec.ref = seg(3);
    unhrec.seq = atol(seg(4).c_str());
    int rc = unhf.write();
    if (!rc || !seq.sgn()) return rc;
    ++seq;
    unhrec.control = seq.pic("##########");
  }
}

int msglog::logUNB(Segment &seg) {
  static PartnerMap pmap;
  if (unbf.keylen() == 0) {
    int rc = unbf;
    if (rc == 0) {
      rc = unbf.open("UNBLOG",BTRDWR + NOKEY);
      if (rc)
	cerr << "UNBLOG: BTAS/X file not found.\n";
    }
    if (rc) return rc;
  }
  unbrec.from = pmap.fromEdiAddr(seg(2),seg(2,1));
  Money seq;
  if (seg(5) == "") {
    char prefix = '0';
    unbf.last(&unbrec.from + Keydata(&prefix,1));
    seq = ++Money(string(unbrec.control).c_str(),"##########");
    seg(5) = seq.pic("##########");
  }
  for (;;) {
    unbrec.control = seg(5);
    unbrec.to = pmap.fromEdiAddr(seg(3),seg(3,1));
    unbrec.timesent = toDateTime(seg(4),seg(4,1));
    unbrec.ack = '0';
    if (seg(9).length())
      unbrec.ack = seg(9)[0];
    int rc = unbf.write();
    if (!rc || !seq.sgn()) return rc;
    ++seq;
    unbrec.control = seq.pic("##########");
  }
}
