#pragma interface
#include <Btcb.h>
#include <vector>
#include "ediseg.h"

// parses a stream into EDI segments
class EdiMsg {
  istream &in;
  char sepstr[2];
  char esc;	// interpret following char literally
  char res;	// EDIFACT reserved control char
public:
  char sep;	// data element seperator
  char segsep;	// segment terminator
  char subsep;	// component data element seperator
  char dec;	// decimal notation
  EdiMsg(istream &);
  // read the ISA segment and set the seperators
  int getISA(const char *,Segment &);
  int getUNB(Segment &);	// read EDIFACT UNB and optional UNA
  // read all remaining segments
  int getseg(Segment &);
  ~EdiMsg();
};

// the base for classes that you write EDI segments to

class EDIsink {
protected:
  EDIsink *errs;	// write err/ack records here
public:
  long cnt;
  EDIsink(EDIsink *e = 0) { errs = e; cnt = 0; }
  void errors(EDIsink *e) { errs = e; }
  virtual int write(Segment &) = 0;
  virtual void error(Segment &fld);	// feedback errs/acks
  virtual ~EDIsink();
};

// Check segment ordering for a transaction set.
// The segments aren't stored anywhere, but write() fails
// if the order is incorrect and appropriate diagnostics
// are fed to error().
// Segments are passed on to another EDIsink if supplied.
class SegmentChecker: public EDIsink {
  SegmentChecker(const SegmentChecker &);
  SegmentChecker&operator=(const SegmentChecker &);
  short type;
  class transet &rec;
  btFile tranf;
  struct loop {
    short beg, end;
    long cnt;
  };
  vector<SegmentChecker::loop>	stk;
  EDIsink *next;
public:
  SegmentChecker(EDIsink * = 0);
  int write(Segment &);
  ~SegmentChecker();
};
