// $Log$
// Revision 1.34  2010/05/12 02:05:30  stuart
// Nits from gcc-4.1.2
//
// Revision 1.33  2010/04/28 19:06:51  stuart
// Nits from 5.4 compiler
//
// Revision 1.32  2010/04/02 04:14:29  stuart
// Convert Array to vector
//
// Revision 1.31  2006/12/11 19:57:29  stuart
// Use century dates for 004000 and *later*.
//
// Revision 1.30  2006/11/17 00:17:30  stuart
// Compile on CentOS-4
//
// Revision 1.29  2005/02/23 21:03:12  stuart
// Fix issue11, deletes last MESSAGE record when nothing to do.
//
// Revision 1.28  2004/02/09 17:28:54  stuart
// Port from libg++ to STL
//
// Revision 1.27  2003/03/25 05:11:48  stuart
// Catch errpost at various spots to improve diagnostics.
//
// Revision 1.26  2003/03/25 01:25:32  stuart
// Fix segcnt check when seg id is EBCDIC (as created by EDX programs)
//
// Revision 1.25  2001/05/02 22:30:46  stuart
// Fix bug where groups are miscounted on error.
//
// Revision 1.24  2001/02/20 22:09:44  stuart
// improve error checking
// update SEND.LOG
//
// Revision 1.23  2000/10/02 18:45:59  ed
// Support century for certain EDI versions.
//
// Revision 1.22  1997/04/18  21:30:23  ed
// Clear segment buffer after validating segments.
// Pass element id to item load to improve diagnostics.
// Fix off by one bug in mandatory element checking.
//
// Revision 1.21  1995/01/18  22:39:11  stuart
// use ANSISink to trim empty elements
//
// Revision 1.20  1994/11/03  17:03:10  ed
// Add additional error messages.
// Fix SE segment count check.
// Fix EDIfile open function.
//
// Revision 1.19  1994/10/27  17:37:14  ed
// allow setting sub-element separator
//
// Revision 1.18  1994/10/20  14:11:18  stuart
// null element handling
//
// Revision 1.17  1994/09/30  15:52:14  stuart
// supply routing source on command line
//
// Revision 1.16  1994/08/03  23:01:07  stuart
// allow setting segment terminator
//
// Revision 1.15  1994/07/28  15:41:12  stuart
// verify command line elem sep
// override message version on command line
// open PARTNERS file *after* help message
//
// Revision 1.14  1994/04/11  18:09:33  stuart
// deleting last MESSAGE record...
//
// Revision 1.13  1994/04/08  20:23:25  stuart
// allow all zero records
// check segment count
//
// Revision 1.12  1994/04/08  16:08:03  stuart
// log message for invalid ST/SE bracket
//
// Revision 1.11  1994/01/31  20:37:29  stuart
// skip incomplete spool files
//

#include <port.h>
#include <stdlib.h>
#include <ctype.h>
#include <ftype.h>
#include <iostream>
#include "edi.h"
#include "edihdr.h"
#include "ansi.h"
#include <errenv.h>

struct config {
  const char *qual;	// routing source qualifier
  const char *src;	// routing source
  const char *sender;	// application sender
  const char *agency;
  const char *version;	// application version
  const char *isaversion;	// application version
};

config cnf = {
  "01", "003423001", "003423001", "X", "003030", "00301"
};

partner partrec;
// construct Keydata separately to work around GNU bug #2504
Keydata prec(&partrec,sizeof partrec);
btFile partf(prec);

class EDIfile: public Btcb, ANSISink {
  string segfile;
  SegDict segdict;
  CodeDict codedict;
  EDIfile(const EDIfile &);
  EDIfile&operator=(const EDIfile &);
  GS grp;
  ISA isa;
  SendLog sendrec;
  btFile grpf, isaf, sndlog, sndgrp, sndmsg;
  long segcnt;
  long tranid;
  short klen;
  char type;	// tran type, 'G' = group passthrough
  bool isa_allocated;	// isa control number allocated
  Segment s;
  vector<string> del;
  void grpend();
  void isasend();
  int load();
public:
  EDIfile(char e = 0x85, char s = '\n', char b = 0x86);
  int segid(string &) const;
  // start message
  void isabeg(const string &qual,const string &recv,char test = 'P');
  // start group
  void grpbeg(FCHAR<4> snd,FCHAR<4> rcv,const string &func);
  // start transaction set
  int open(const string &dir,const string &name);	
  int send(bool wflg = true);	// validate, if wflg, send record to os
  void close();	// end transaction set
  void isaend(char purge = 0);	// end message, optionally delete spool files
  ~EDIfile();
};

EDIfile::EDIfile(char e, char s, char b):
  grpf("GROUP",BTRDWR,Keydata(&grp,sizeof grp)),
  isaf("MESSAGE",BTRDWR,Keydata(&isa,sizeof isa)),
  sndlog(Keydata(&sendrec,sizeof sendrec)),
  sndgrp(&sendrec.grp + &sendrec.no + &sendrec.mod + &sendrec.file
      + &sendrec.files + &sendrec.seq + &sendrec.set + &sendrec.type),
  sndmsg(&sendrec.no + &sendrec.mod + &sendrec.file
      + &sendrec.files + &sendrec.seq + &sendrec.set + &sendrec.type),
  codedict("/edx/AIRCO/EDI/CODE.MAST"),
  segfile(string("/bms/edi/") + cnf.version + ".seg"),
  ANSISink(e,s,b),
  segdict(segfile.c_str())
{
  klen = 0;
  isa_allocated = false;
  if (sndlog.open("../SND.LOG",BTRDWR) == 0
    && sndmsg.open("../SND.LOG.MSG",BTRDWR) == 0
    && sndgrp.open("../SND.LOG.GRP",BTRDWR) == 0) {
    sndmsg.relate(&sndlog);
    sndmsg.relate(&sndgrp);
  }
  else {
    cerr << "Could not open ../SND.LOG*\n";
  }
}

EDIfile::~EDIfile() {
  isaend();
}

struct QualAddr {
  string addr, qual;
  QualAddr(string str);
};

QualAddr::QualAddr(string str) {
  size_t pos = str.find('-');
  if (pos != string::npos) {
    qual = str.substr(0,pos);
    addr = str.substr(pos + 1);
    return;
  }
  pos = str.find(':');
  if (pos != string::npos) {
    qual = str.substr(pos + 1);
    addr = str.substr(0,pos);
    return;
  }
  addr = str;
}

void EDIfile::isabeg(const string &sndr, const string &recv,char test) {
  isaend();
  QualAddr addr(sndr);
  isa.sendqual= addr.qual;
  isa.sender	= addr.addr;
  isa.standard= 'U';
  isa_allocated = false;
  if (isaf.last(&isa.sendqual + &isa.sender + Keydata(&isa.standard,1)))
    isa.ctlno = 0;
  isa.authqual= "00";	// no authorization info
  isa.secqual	= "00";	// no security info
  QualAddr rcvr(recv);
  isa.recvqual= rcvr.qual;
  isa.receiver= rcvr.addr;
  isa.time	= time(0);
  isa.version	= cnf.isaversion;
  isa.ackreq	= '0';
  isa.test	= test;
  isa.grps	= 0;
}

void EDIfile::isasend() {
  do
    ++isa.ctlno;
  while (isaf.write());
  isa_allocated = true;
  time_t now = isa.time;
  tm *t = localtime(&now);
  cout	<< "ISA" << elemsep
	<< isa.authqual << elemsep << trim(isa.authinfo,10,10) << elemsep
	<< isa.secqual  << elemsep << trim(isa.secinfo,10,10)  << elemsep
	<< isa.sendqual << elemsep << trim(isa.sender,15,15)   << elemsep
	<< isa.recvqual << elemsep << trim(isa.receiver,15,15) << elemsep
	<< tm_mask(t,"YYMMDD") << elemsep << tm_mask(t,"hhmm") << elemsep
	<< isa.standard << elemsep
	<< trim(isa.version,5,5) << elemsep
	<< isa.ctlno.pic("#########") << elemsep
	<< isa.ackreq << elemsep << isa.test << elemsep
	<< subsep << segsep;
}

void EDIfile::grpbeg(FCHAR<4> snd,FCHAR<4> rcv,const string &func) {
  grpend();
  if (!isa.grps)
    isasend();
  ++isa.grps;
  if (type == 'G') return;	// self contained group records
  time_t now = time(0);
  tm *t = localtime(&now);
  if (partf.readeq(&snd)) {
    cerr << "Partner id " << snd << " not found.\n";
  }
  grp.sndr = partrec.addr;
  if (grpf.last(&grp.sndr))
    grp.ctlno = 0;
  grp.func = func;
  if (partf.readeq(&rcv)) {
    cerr << "Partner id " << rcv << " not found.\n";
  }
  grp.rcvr = partrec.addr;
  grp.agency = cnf.agency;
  grp.version = cnf.version;
  grp.time = now;
  grp.sets = 0;
  do
    ++grp.ctlno;
  while (grpf.write());

  const char *datepic;
  if (CodeDict::century == true)
    datepic = "YYYYMMDD";
  else
    datepic = "YYMMDD";

  cout	<< "GS" << elemsep
  	<< grp.func << elemsep
	<< grp.sndr << elemsep
	<< grp.rcvr << elemsep
	<< tm_mask(t,datepic) << elemsep
	<< tm_mask(t,"hhmm") << elemsep
	<< grp.ctlno << elemsep
	<< grp.agency << elemsep
	<< grp.version << segsep;
  cout.flush();
}

int EDIfile::open(const string &dir,const string &fname) {
  close();
  segcnt = 0;
  tranid = 0;
  string name = dir + '/' + fname;
  int rc = Btcb::open(fname.c_str(),BTRDONLY + BTDIROK);
  if (rc == 0) {
    string seg1st, seglast, tran1st;
    int pos = 0;
    klen = bt_klen(b->lbuf);
    s.clear();
    rc = first();
    if (!rc && !(rc = load())) {
      seg1st = s[0];
      tran1st = s[2];
    }
    s.clear();
    if (rc || (rc = last())  || !(pos = segid(seglast))
	|| seglast != seg1st.substr(0,1) + 'E' || load() || tran1st != s[2]) {
      cerr << name << ": invalid or incomplete\n";
      s.clear();
      return rc ? rc : -1;
    }
    s.clear();
    // if SE/ST message, validate all segments without transmitting first
    // (we don't validate GS/GE messages)
    type = seg1st[0];
    if (type == 'S') {
      segcnt = ldlong(b->lbuf + pos + 1);
      catch (rc)
      for (rc = first(); rc == 0; rc = next()) {
	if (send(false)) { 	// validate segment
	  segid(seglast);
	  cerr << name << ": " << seglast << " - invalid segment\n";
	  return -1;
	}
      }
      rc = 0;
      envend
      if (rc > 0) {
	cerr << "Btas error " << rc << " validating segments." << endl;
	cerr << "Last segment: " << seglast << endl;
	return -1;
      }
      if (segcnt) {
	cerr << "Invalid segment count in SE segment\n";
	return -1;
      }
      tranid = atol(tran1st.c_str());
      //cerr << "Tran id " << tranid << '\n';
    }
    else {
      cerr << "Funky type " << type << '\n';
      tranid = 0;
      segcnt = 0;
    }
    del.push_back(name);
    return 0;
  }
  if (rc == 210)
    cerr << "Skipping subdirectory `" << name << "'\n";
  else
    cerr << name << ": BTAS error " << rc << endl;
  return rc;
}

void EDIfile::close() {
  if (klen) {
    klen = 0;
    Btcb::close();
    if (tranid > 0)
      ++grp.sets;
    if (grp.ctlno) {
      sendrec.no = tranid;
      int rc = sndmsg.readeq(&sendrec.no);
      if (rc == 0) {
	sendrec.grp = grp.ctlno;
	sendrec.grptm = grp.time;
	sendrec.stat = "S";
	catch (rc)
	  sndmsg.rewrite();
	  rc = 0;
	envend
	if (rc)
	  cerr << "BTAS error " << rc << " updating tran " << tranid << endl;
      }
    }
  }
}

void EDIfile::grpend() {
  if (grp.ctlno) {
    grpf.rewrite();
    cout	<< "GE" << elemsep
	<< grp.sets << elemsep
	<< grp.ctlno << segsep;
    cout.flush();
    grp.ctlno = 0;
  }
}

void EDIfile::isaend(char purge) {
  close();
  grpend();
  if (isa.grps) {
    isaf.rewrite();
    cout	<< "IEA" << elemsep
	<< isa.grps << elemsep << isa.ctlno.pic("#########") << segsep;
    isa.grps = 0;
    isa.ctlno = 0;
    if (purge) {
      for (size_t i = 0; i < del.size(); ++i) {
	btkill(del[i].c_str());
	cerr << "btkill(" << del[i] << ")\n";
      }
    }
  }
  else if (isa.ctlno && isa_allocated)
    isaf.del();
  isa_allocated = false;
  del.clear();
}

int EDIfile::segid(string &seg) const {
  int flen = 3;
  char buf[4];
  /* if (len() <= klen + flen)		// short record */
    /* return 0; */
  if (lbuf()[klen] & 0x80)
    ldebstr(buf,(unsigned char *)lbuf()+klen,flen);
  else
    ldchar(lbuf()+klen,flen,buf);
  seg = buf;
  return klen + flen;
}

int EDIfile::send(bool wflg) {
  int rc = load();
  if (wflg)
    ANSISink::write(s);
  else
    s.clear();	// throw segment away
  --segcnt;
  return rc;
}

/** Load segment from BTAS buffer. */
int EDIfile::load() {
  int rc = 0;
  string seg;
  const SegDict::Tag *p = 0;
  const char *req = 0;
  int pos = segid(seg);
  catch (rc)
  if (pos) {
    p = segdict[seg.c_str()];
    req = segdict.req(seg.c_str());
  }
  else
    seg = "???";
  if (!pos || !p) {
    cerr << "Segment " << seg << " is invalid\n";
    return -1;
  }
  s.add(seg);
  char elemid[6];
  strncpy(elemid,seg.c_str(),3);
  elemid[5] = 0;
  for (int i = 0; p[i]; ) {
    EDIitem &c = codedict[p[i++]];
    int flen = c.len();
    if (!flen) continue;
    elemid[3] = '0' + i/10;
    elemid[4] = '0' + i%10;
    pos = c.align(pos);
    if (pos >= len())
      pos = len();
    if (pos + flen > len())
      memset(lbuf() + len(),0,pos + flen - len());
    // NOTE: first fld is 1 here
    bool mand = req && req[i - 1] == 'M';
    const char *buf = lbuf() + pos;
    s.add(c.load(buf,mand,elemid));
    const char *msg = 0;
    if (mand || !c.isnull(buf))
      msg = c.validate(buf);
    if (msg) {
      cerr << "Element " << elemid << " is invalid\n";
      rc = -1;
    }
    pos += flen;
  }
  envend
  if (rc > 0)
    cerr << "Btas error " << rc << " loading segment " << seg << '\n';
  return rc;
}

GOTO usage() {
  cerr << "	$Id$\n\
Usage:	edixmit [-k] [-t] [-ec] [-vnum] routing-src routing-dest id [id ...]\n\
	routing-dest	ISA qual-destination, e.g. 01-555463785\n\
			or destination:qual, e.g. 555463785:01\n\
	-k		keep spool files\n\
	-t		generate test message\n\
	-ec		set element separator to c\n\
	-en		set element separator to ascii code n\n\
	-sc		set segment terminator to c\n\
	-sn		set segment terminator to ascii code n\n\
	-vnum		override message version to 'num'\n\
	-inum		override isa version to 'num'\n\
\n\
Files:	BTASDIR				BTAS Spool Directory\n\
	/bms/edi/*.seg			UNIX Segment Table by message version\n\
	/edx/AIRCO/EDI/CODE.MAST	BTAS Code Dictionary\n\
	MESSAGE				BTAS Message log\n\
	GROUP				BTAS Group log\n\
	PARTNERS			BTAS EDI Addresses\n\
	../SND.LOG			SEND LOG\n\
";
  exit(1);
}

struct options {
  int keep:1;
  char test;
  char elemsep, segsep, subsep;
} opt = {
  0, 'P', 0x1d, 0x1c, 0x1f		// Kleinschmidt agreed on values
};

int main(int argc,char **argv) {
  // parse options
  while (argc > 1 && *argv[1] == '-') {
    char c;
    --argc, ++argv;
    while (*++argv[0]) {
      switch (*argv[0]) {
      case 'k':	// keep spool files
	opt.keep = 1;
	continue;
      case 't':	// test message
	opt.test = 'T';
	continue;
      case 's':	// segment terminator
	c = *++argv[0];
	if (isdigit(c))
	  opt.segsep = atoi(*argv);
	else if (c && c != ' ')
	  opt.segsep = c;
	break;
      case 'e':	// element separator
	c = *++argv[0];
	if (isdigit(c))
	  opt.elemsep = atoi(*argv);
	else if (c && c != ' ')
	  opt.elemsep = c;
	break;
      case 'b':	// sub-element separator
	c = *++argv[0];
	if (isdigit(c))
	  opt.subsep = atoi(*argv);
	else if (c && c != ' ')
	  opt.subsep = c;
	break;
      case 'v':
	cnf.version = ++argv[0];
	break;
      case 'i':
	cnf.isaversion = ++argv[0];
	break;
      default:
	cerr << "Bad switch '" << *argv[0] << "'\n";
	usage();
      }
      break;
    }
  }
  if (argc < 4)
    usage();

  /** Use century dates when version is 004010 */
  if (cnf.version[2] >= '4') CodeDict::century = true;
  partf.open("PARTNERS",BTRDONLY);
  // transmit messages
  EDIfile f(opt.elemsep,opt.segsep,opt.subsep);
  f.isabeg(argv[1],argv[2],opt.test);
  string grpnam;		// grouping part of filename
  for (int i = 3; i < argc; ++i) {
    string id = argv[i];
    string pat = id + "/[A-Z]*";
    int rc;
    catch (rc)
    for (Btfind ff(pat.c_str()); !ff; ff.next()) {
      string spooldir(ff.dirname());
      string spoolname(ff.name());
      catch (rc)
      if (f.open(spooldir,spoolname) == 0) {
	cerr << "Transmitting " << id + '/' + spoolname << endl;
	vector<string> part;
	split(spoolname,part,3,string(1,'.'));
	if (id + part[0] + part[1] != grpnam) {	// new group
	  grpnam = id + part[0] + part[1];
	  f.grpbeg(part[0],id,part[1].substr(0,2));
	}
	catch (rc)
	  for (rc = f.first(); rc == 0; rc = f.next())
	    f.send();
	  rc = 0;
	envend
	if (rc)
	  cerr << "Btas error " << rc << " reading spool file" << endl;
      }
      rc = 0;
      envend
      if (rc)
	cerr << "Btas error " << rc << " opening spool file: "
	     << spooldir << '/' << spoolname << endl;
    }
    envend
    if (rc)
      cerr << "Btas error " << rc << " reading spool directory\n";
  }
  if (!opt.keep)
    f.isaend(1);
  return 0;
}
